//*****************************************************************************
//              \ /
// No Copyright (c) 2003
//              / \
// No rights reserved. This material contains published,
// non-copyrighted work, which includes non-confidential and
// non-proprietary information of Kurt Santel.
//
// Source: CodeWar.cpp
//
// Description: Server-side component of CodeWar
//
//*****************************************************************************

#include "CodeWar.h"

#include "../CodeWarCommon/CodeWarCommands.h"
#include "Arena.h"

////////////////////////////////////////////////////////////////////
// TRACE

#if 0
//#if _DEBUG

#include <time.h>
void TRACE(char * s)
{
    FILE * f = ::fopen( "C:/CodeWar.log", "a" );
    ::fwrite( "----------", sizeof(char), 10, f );
    time_t theTime;
    time( &theTime );
    char timeBuffer[100];
    sprintf( timeBuffer, "%s", ctime( &theTime ) );
    ::fwrite( timeBuffer, sizeof(char), ::strlen(timeBuffer), f );
    ::fwrite( s, sizeof(char), ::strlen(s), f );
    ::fclose( f );
}

#else

void TRACE(char * s)
{
    // Do nothing
}

#endif

////////////////////////////////////////////////////////////////////
// CodeWar

CodeWar::CodeWar(void)
: m_ListeningThread( NULL )
, m_CleanupThread( NULL )
, m_ControlThread( NULL )
, m_SlaveThreadMutex( NULL )
, m_Arena( NULL )
{
}

CodeWar::~CodeWar(void)
{
    TRACE("CodeWar::~CodeWar\n");

    // Stop and delete the listening thread
    if( m_ListeningThread )
    {
        m_ListeningThread->Stop();
        m_ListeningThread->WaitUntilTerminate();

        delete m_ListeningThread;
        m_ListeningThread = NULL;
    }

    // Stop and delete the cleanup thread
    if( m_CleanupThread )
    {
        m_CleanupThread->Stop();
        m_CleanupThread->WaitUntilTerminate();

        delete m_CleanupThread;
        m_CleanupThread = NULL;
    }

    // Stop and delete the listening thread
    if( m_ControlThread )
    {
        m_ControlThread->Stop();
        m_ControlThread->WaitUntilTerminate();

        delete m_ControlThread;
        m_ControlThread = NULL;
    }

    // Stop and delete all the slave threads
    SlaveThreadType::const_iterator end = m_SlaveThreads.end();
    SlaveThreadType::iterator it = m_SlaveThreads.begin();

    while( end != it )
    {
        SlaveThread * slave = *it;

        slave->Stop();
        slave->WaitUntilTerminate();

        delete slave;

        ++it;
    }

    m_SlaveThreads.clear();

    // Clean up the mutex
    if( m_SlaveThreadMutex )
    {
        ::CloseHandle( m_SlaveThreadMutex );
        m_SlaveThreadMutex = NULL;
    }

    if( m_Arena )
    {
        delete m_Arena;
        m_Arena = NULL;
    }
}

bool CodeWar::initialize(XType width, YType height, unsigned short port)
{
    if( NULL == m_SlaveThreadMutex )
    {
        m_SlaveThreadMutex = ::CreateMutex( NULL, FALSE, NULL );
    }

    if( NULL == m_SlaveThreadMutex )
    {
        return( false );
    }

    // Create the ListeningThread
    if( m_ListeningThread )
    {
        m_ListeningThread->Stop();
        m_ListeningThread->WaitUntilTerminate();

        delete m_ListeningThread;
        m_ListeningThread = NULL;
    }

    m_ListeningThread = new ListeningThread( this, port );

    if( NULL == m_ListeningThread )
    {
        return( false );
    }

    m_ListeningThread->Start();

    // Create the CleanupThread
    if( m_CleanupThread )
    {
        m_CleanupThread->Stop();
        m_CleanupThread->WaitUntilTerminate();

        delete m_CleanupThread;
        m_CleanupThread = NULL;
    }

    m_CleanupThread = new CleanupThread( this );

    if( NULL == m_CleanupThread )
    {
        return( false );
    }

    m_CleanupThread->Start();

    // Create the ControlThread
    if( m_ControlThread )
    {
        m_ControlThread->Stop();
        m_ControlThread->WaitUntilTerminate();

        delete m_ControlThread;
        m_ControlThread = NULL;
    }

    m_ControlThread = new ControlThread( this );

    if( NULL == m_ControlThread )
    {
        return( false );
    }

    m_ControlThread->Start();

    if( NULL == m_Arena )
    {
        m_Arena = new Arena(width, height);

        if( NULL == m_Arena )
        {
            return( false );
        }
    }

    return( true );
}

void CodeWar::acceptConnection(SOCKET s)
{
    int Error;

    TRACE("CodeWar::acceptConnection::WaitForSingleObject\n");
    ::WaitForSingleObject( m_SlaveThreadMutex, INFINITE );

    // Create a new SlaveThread
    SlaveThread * slave = new SlaveThread( this, s );

    // Add it to m_SlaveThreads
    if( slave )
    {
        slave->Start();

        m_SlaveThreads.push_back( slave );
    }
    else
    {
        if( SOCKET_ERROR == ::shutdown( s, 2 ) )
        {
            Error = ::WSAGetLastError();
        }

        if( SOCKET_ERROR == ::closesocket( s ) )
        {
            Error = ::WSAGetLastError();
        }
    }

    TRACE("CodeWar::acceptConnection::ReleaseMutex\n");
    ::ReleaseMutex( m_SlaveThreadMutex );
}

////////////////////////////////////////////////////////////////////
// CodeWar::ListeningThread member functions

CodeWar::ListeningThread::ListeningThread(CodeWar * cw, unsigned short port)
: CThread()
, m_bStop( false )
, m_Port( port )
, m_CodeWar( cw )
, m_ListeningSocket( INVALID_SOCKET )
{
}

void CodeWar::ListeningThread::OnRunning(void)
{
    int Error;

    while( !m_bStop )
    {
        sockaddr saTemp;

        // Block until a client connects
        SOCKET client = ::accept( m_ListeningSocket, &saTemp, 0 );

        if( INVALID_SOCKET != client )
        {
            // Set Receive Time out
            int timeout = 10000;

            if( SOCKET_ERROR != ::setsockopt( client,
                                              SOL_SOCKET,
                                              SO_RCVTIMEO,
                                              (char*)&timeout,
                                              sizeof(int) ) )
            {
                if( m_CodeWar )
                {
                    m_CodeWar->acceptConnection( client );
                }
            }
        }
        else
        {
            Error = ::WSAGetLastError();
        }
    }
}

void CodeWar::ListeningThread::Stop(void)
{
    int Error;

    if( !m_bStop )
    {
        m_bStop = true;

        if( INVALID_SOCKET != m_ListeningSocket )
        {
            if( SOCKET_ERROR == ::shutdown( m_ListeningSocket, 2 ) )
            {
                Error = ::WSAGetLastError();
            }

            if( SOCKET_ERROR == ::closesocket( m_ListeningSocket ) )
            {
                Error = ::WSAGetLastError();
            }
        }
    }
}

void CodeWar::ListeningThread::OnInitInstance()
{
    int Error;

    m_ListeningSocket = ::socket( AF_INET, SOCK_STREAM, 0 );

    if( INVALID_SOCKET != m_ListeningSocket )
    {
        SOCKADDR_IN saTemp;

        saTemp.sin_family = AF_INET;
        saTemp.sin_port = ::htons( m_Port );
        saTemp.sin_addr.s_addr= INADDR_ANY;

        if( SOCKET_ERROR == ::bind( m_ListeningSocket, (PSOCKADDR)&saTemp, sizeof(saTemp) ) )
        {
            Error = ::WSAGetLastError();
        }

        if( SOCKET_ERROR == ::listen( m_ListeningSocket, SOMAXCONN ) )
        {
            Error = ::WSAGetLastError();
        }
    }
    else
    {
        Error = ::WSAGetLastError();
    }
}

////////////////////////////////////////////////////////////////////
// CodeWar::SlaveThread member functions

CodeWar::SlaveThread::SlaveThread(CodeWar * cw, SOCKET client)
: CThread()
, m_CodeWar( cw )
, m_bStop( false )
, m_ClientSocket( client )
{
}

void CodeWar::SlaveThread::OnRunning(void)
{
    int Error;                  // Only needed for debug purposes
    BasicCommand cmd;           // Every command starts with a header
    int BytesRead;

    while( !m_bStop )
    {
        // Take a peek at the command that will/has arrive/arrived
        BytesRead = ::recv( m_ClientSocket, (char *)&cmd, sizeof(cmd), MSG_PEEK );

        // Check if the socket communication succeeded
        if( (SOCKET_ERROR == BytesRead) || (0 == BytesRead) )
        {
            Error = ::WSAGetLastError();
            Stop();
        }

        // Check the version of the command
        if( theVersion != cmd.m_version )
        {
            Stop();
        }

        // A new client needs to register
        if( !m_RobotProperties.m_registered )
        {
            if( cmdRegister == cmd.m_command )
            {
                registerClient();
            }
            else
            {
                Stop();
            }
        }
        else        // Client is already registered
        {
            doTurn( cmd.m_command );
        }
    }
}

void CodeWar::SlaveThread::Stop(void)
{
    int Error;

    if( !m_bStop )
    {
        m_bStop = true;

        if( INVALID_SOCKET != m_ClientSocket )
        {
            if( SOCKET_ERROR == ::shutdown( m_ClientSocket, 2 ) )
            {
                Error = ::WSAGetLastError();
            }

            if( SOCKET_ERROR == ::closesocket( m_ClientSocket ) )
            {
                Error = ::WSAGetLastError();
            }
        }
    }
}

bool CodeWar::SlaveThread::Stopped(void)
{
    return( m_bStop );
}

void CodeWar::SlaveThread::registerClient(void)
{
    int Error;
    registerCommand rc;

    // Receive the complete command
    int BytesRead = ::recv( m_ClientSocket, (char *)&rc, sizeof(rc), 0 );

    // Check if the socket communication succeeded
    if( (SOCKET_ERROR == BytesRead) || (0 == BytesRead) )
    {
        Error = ::WSAGetLastError();
        Stop();
    }

    // Fill in the properties
    m_RobotProperties.m_hasTurn = false;
    m_RobotProperties.m_hadTurn = true;
    m_RobotProperties.m_ID = this->m_dwThreadID;    // Use the ThreadID
    m_RobotProperties.m_attack = rc.m_attack;
    m_RobotProperties.m_defense = rc.m_defense;
    m_RobotProperties.m_food = rc.m_food;

    //******************************************************
    // Assign a position in the arena to the robot
    //******************************************************
    m_CodeWar->m_Arena->GetRandomPosition( &m_RobotProperties.m_position );

    m_RobotProperties.m_registered = true;
}

void CodeWar::SlaveThread::doTurn(CommandType & cmdType)
{
    int Error;                  // Only needed for debug purposes
    int BytesRead;
    bool succeeded;

    // All possible commands that can arrive
    moveForwardCommand mf;
    idleCommand idle;

    // Check some conditions before continuing
    if( !m_RobotProperties.checkPreConditions() )
    {
        Stop();
        return;
    }

    // Receive the complete command
    switch( cmdType )
    {
    case cmdIdle:
        BytesRead = ::recv( m_ClientSocket, (char *)&idle, sizeof(idle), 0 );
        break;
    case cmdRegister:       // Invalid command
        Stop();
        return;
    case cmdPoll:           // Invalid command
        Stop();
        return;
    case cmdMoveForward:
        BytesRead = ::recv( m_ClientSocket, (char *)&mf, sizeof(mf), 0 );
        break;
    }

    // Check if the socket communication succeeded
    if( (SOCKET_ERROR == BytesRead) || (0 == BytesRead) )
    {
        Error = ::WSAGetLastError();
        Stop();
        return;
    }

    switch( cmdType )
    {
    case cmdIdle:
        succeeded = m_RobotProperties.doTurn( idle );
        break;
    case cmdMoveForward:
        succeeded = m_RobotProperties.doTurn( mf );
        break;
    }

    if( !succeeded || !m_RobotProperties.checkPostConditions() )
    {
        Stop();
        return;
    }
}

////////////////////////////////////////////////////////////////////
// CodeWar::CleanupThread member functions

CodeWar::CleanupThread::CleanupThread(CodeWar * cw)
: CThread()
, m_bStop( false )
, m_CodeWar( cw )
{
}

void CodeWar::CleanupThread::OnRunning(void)
{
    DWORD milliseconds;

    while( !m_bStop )
    {
        milliseconds = 2000;                // Sleep for 2 seconds except if
                                            // a slave thread gets deleted.
        if( m_CodeWar )
        {
            TRACE("CleanupThread::WaitForSingleObject\n");
            ::WaitForSingleObject( m_CodeWar->m_SlaveThreadMutex, INFINITE );

            SlaveThreadType::const_iterator end = m_CodeWar->m_SlaveThreads.end();
            SlaveThreadType::iterator it = m_CodeWar->m_SlaveThreads.begin();

            while( end != it )
            {
                SlaveThread * slave = *it;

                if( slave->Stopped() )
                {
                    slave->WaitUntilTerminate();

                    delete slave;           // DDDDIIIIIEEEEEEEEEE !!!!

                    m_CodeWar->m_SlaveThreads.erase( it );

                    milliseconds = 0;       // Do not sleep! Work, baby, work!

                    TRACE("ControlThread::SlaveThread has been deleted\n");

                    break;
                }

                ++it;
            }

            TRACE("CleanupThread::ReleaseMutex\n");
            ::ReleaseMutex( m_CodeWar->m_SlaveThreadMutex );
        }

        ::Sleep( milliseconds );
    }
}

void CodeWar::CleanupThread::Stop(void)
{
    m_bStop = true;
}

////////////////////////////////////////////////////////////////////
// CodeWar::ControlThread member functions

CodeWar::ControlThread::ControlThread(CodeWar * cw)
: CThread()
, m_bStop( false )
, m_CodeWar( cw )
{
}

void CodeWar::ControlThread::OnRunning(void)
{
    size_t i = 0;               // current position in m_SlaveThreads
    SlaveThread * slave = NULL;

    //**********************************************************
    // This turn based algorithm needs to be checked
    // What if threads are closed?  Is the order still correct?
    // Do some clients get more turns?
    //**********************************************************

    //**********************************************************
    // New proposal:
    // walk thru the list with SlaveThreads
    // and check which one needs to send something
    // send it, and get the next SlaveThread ready
    //**********************************************************

    //**********************************************************
    // TODO:
    // Still need to check what happens if a client stops sending commands.
    // This thread will probably loop doing nothing.
    // Some extra security (timeout mechanism) needs to build in.
    //**********************************************************

    while( !m_bStop )
    {
        if( m_CodeWar )
        {
            TRACE("ControlThread::WaitForSingleObject\n");
            ::WaitForSingleObject( m_CodeWar->m_SlaveThreadMutex, INFINITE );

            size_t size = m_CodeWar->m_SlaveThreads.size();

            if( i < size )
            {
                slave = m_CodeWar->m_SlaveThreads.at( i );

                if( slave )
                {
                    if( 1 == size )                             // If there is only one
                    {                                           // client, let it play.
                        slave->m_RobotProperties.m_hadTurn = false;
                    }
                    else
                    {
                        slave = slave;  // debug purpose: set a break point here
                    }

                    if( slave->m_RobotProperties.m_hadTurn ||    // The client sent a command
                        !slave->m_RobotProperties.m_registered ) //   or is not yet registered
                    {                                           
                        ++i;                                    // Thus the next in line
                                                                // can play.
                        if( i >= size )
                        {
                            i = 0;
                        }

                        slave = m_CodeWar->m_SlaveThreads.at( i );

                        if( slave )
                        {
                            slave->m_RobotProperties.m_hadTurn = false;
                        }
                    }
                    else
                    {
                        if( !slave->m_RobotProperties.m_hasTurn )
                        {
                            sendCommands( slave );
                        }
                    }
                }
            }
            else
            {
                i = 0;
            }

            TRACE("ControlThread::ReleaseMutex\n");
            ::ReleaseMutex( m_CodeWar->m_SlaveThreadMutex );

            // Since controlling all these clients is exhausting, catch a small 'z'
            ::Sleep( 10 );
        }
    }
}

void CodeWar::ControlThread::Stop(void)
{
    m_bStop = true;
}

void CodeWar::ControlThread::sendCommands(SlaveThread * slave)
{
    int Error;
    int BytesSent;

    // Check if events need to be sent
    if( slave->m_RobotProperties.m_wallHit )
    {
        onHitWallCommand onHitWallCmd;

        // Send the command to the client
        BytesSent = ::send( slave->m_ClientSocket, (char *)&onHitWallCmd, sizeof(onHitWallCmd), 0 );

        if( SOCKET_ERROR == BytesSent )
        {
            Error = ::WSAGetLastError();
            Stop();
        }
    }

    // Prepare the poll command
    StatusInfo info( slave->m_RobotProperties.m_food,
                     slave->m_RobotProperties.m_life,
                     slave->m_RobotProperties.m_heading );

    infoCommand infoCmd( slave->m_RobotProperties.m_ID, info );

    // Send the command to the client
    BytesSent = ::send( slave->m_ClientSocket, (char *)&infoCmd, sizeof(infoCmd), 0 );

    if( SOCKET_ERROR == BytesSent )
    {
        Error = ::WSAGetLastError();
        Stop();
    }

    pollCommand pollCmd;

    // Send the command to the client
    BytesSent = ::send( slave->m_ClientSocket, (char *)&pollCmd, sizeof(pollCmd), 0 );

    if( SOCKET_ERROR == BytesSent )
    {
        Error = ::WSAGetLastError();
        Stop();
    }

    slave->m_RobotProperties.m_hasTurn = true;
}
